unit libSystem;
// Purpose : Implement the system functions
// Created : YR - November 2000
interface

uses SysUtils, Windows,WinSvc,JclSYSInfo, Classes;

	// Define a circular buffer to store CPU's values
	type TCircularBuffer = class(TObject)
		private
	   	values : array of integer;
	   	curidx : integer;
         maxidx : integer;
         function GetValue(nOffset: integer) : integer;
   	public
			constructor Create(nSize : integer);
			procedure Push(value: integer);
         function GetAverage(nRange : integer) : integer;
	end;

	procedure FillArray(var av : array of variant; v : variant);
	procedure WriteToEventLog(sMsg : String);
	function Get_ComputerName : string;
   function GetFreeSpace(pDrive: pchar) : int64;
   function GetDiskSize(pDrive: pchar) : int64;
   function IsNumber(s: string) : boolean;
	function ServiceRunning(sMachine, sService : string ) : boolean;
	function ServiceStopped(sMachine,sService : string ) : boolean;
	Function ProcessRunning(sName: string) : boolean;

implementation

	constructor TCircularBuffer.Create(nSize : integer);
	// Purpose	: Create buffer dynamically
	// Created	: YR - 21.11.2000
	var i: integer;
	begin
		inherited Create;
		maxidx := 0;
		maxidx := nSize;
		SetLength(values, maxidx);
	   // Initialize values to -1
	   for i:=0 to maxidx-1 do
	   	values[i]:=-1;
	end;

	procedure TCircularBuffer.Push(value: integer);
	// Purpose	: Push function: Used to add a value to the buffer
	// Created	: YR - 21.11.2000
	begin
		if curidx>=maxidx then curidx := 0;
	   values[curidx]:=value;
		inc(curidx);
	end;

	function TCircularBuffer.GetValue(nOffset: integer) : integer;
	// Purpose	: Get value corresponding to the last entered value - nOffset;
	// Created	: YR - 21.11.2000
	// Return	: Actual value or
	//				  -2 if nOffset bigger than array size or
	//				  -1 if value is undefined
	begin
		if nOffset>length(values) then
			// nOffset is bigger than the array size
	   	GetValue:=-2
		else
		begin
			if (curidx-1) >= nOffset then
				// actual value is comming from 0 to current index
				GetValue := values[(curidx-1)-nOffset]
		   else
				// actual value is comming from the end of the buffer
	      	GetValue := values[length(values) + (curidx-1) - nOffset];
	   end;
	end;


	function TCircularBuffer.GetAverage(nRange : integer) : integer;
	// Purpose	: Get an average value on the <nRange> last values
	// Created	: YR - 21.11.2000
	var
	 	nOff: integer;
		nSum : integer;
		nNbTerm : integer;
	   nCurValue : integer;
	begin
		nOff:= 0;
	   nSum := 0;
	   nNBTerm := 0;
	   while (nOff<nRange) do
	   begin
			nCurValue := GetValue(nOff);
	   	if nCurValue >= 0 then
	      begin
	      	inc(nSum,nCurValue);
	         inc(nNbTerm);
	      end;
	      inc(nOff);
	   end;
	   GetAverage := nSum div nNbTerm;
	end;

	procedure WriteToLog(sLine : String);
	// Write an line in the log file
	// Created	: YR - 21.11.2000
	var
	        fLOG : text;
	begin
	        AssignFile(fLOG,'NSClient.log');
	        append(fLOG);
	        Writeln(fLOG,sLine);
	        flush(fLOG);
	        closefile(fLOG);
	end;

	function Get_ComputerName : string;
	// Purpose	: Return the computer name of the local machine
	// Created	: YR - 21.11.2000
	var
	  ComputerName : string;
	  ComputerNameLen : DWORD;
	begin
   	SetLength(ComputerName, MAX_COMPUTERNAME_LENGTH + 1);
	   GetComputerName(PChar(ComputerName), ComputerNameLen);
		SetLength(ComputerName, ComputerNameLen);
		result := ComputerName;
	end;

	procedure FillArray(var av : array of variant; v : variant);
	// Purpose	: Fill the array with the v value
	// Created	: YR - 22.11.2000
	var
		i: integer;
	begin
		for i:=0 to length(av)-1 do av[i]:=v;
	end;

   function IsNumber(s: string) : boolean;
	// Purpose	: Return true if the string can be converted to a number
	// Created	: YR - 22.11.2000
	var
   	i: integer;
	begin
		IsNumber := true;
   	for i:=0 to length(s) do begin
      	case ord(s[i]) of
         	1..47,58..255:	begin
              	IsNumber:=false;
               Break;
            end;
         end;
      end;
   end;

   function GetFreeSpace(pDrive: pchar) : int64;
	// Purpose	: Return the free disk space of a drive
	// Created	: YR - 23.11.2000
   var
   	nDriveNumber: byte;
   begin
		nDriveNumber := ord(pDrive[0])-64;
		case nDriveNumber of
      	3..26: GetFreeSpace:=DiskFree(nDriveNumber);	// we don't want to allow queries on A & B drive
         else GetFreeSpace:=-1;
      end;
   end;

   function GetDiskSize(pDrive: pchar) : int64;
	// Purpose	: Return the total disk space of a drive
	// Created	: YR - 23.11.2000
   var
   	nDriveNumber: byte;
   begin
		nDriveNumber := ord(pDrive[0])-64;
		case nDriveNumber of
      	3..26: GetDiskSize:=DiskSize(nDriveNumber);	// we don't want to allow query on A & B drive
         else GetDiskSize:=-1;
      end;
   end;

	function ServiceGetStatus(sMachine,sService : string ) : Integer;
	// Purpose : get service status
	// return status code if successful
	// -1 if not
	//
	// return codes:
	// 	  SERVICE_STOPPED
	//   SERVICE_RUNNING
	//   SERVICE_PAUSED
	//
	// following return codes are used to indicate that
	// the service is in the middle of getting to one
	// of the above states:
	// 	SERVICE_START_PENDING
	// 	SERVICE_STOP_PENDING
	// 	SERVICE_CONTINUE_PENDING
	//    SERVICE_PAUSE_PENDING
	//
	// sMachine: machine name, ie: \\SERVER
	// empty = local machine
	//
	// sService:  service name, ie: Alerter

	// Created : http://www.chami.com/tips/delphi/031498D.html
	var
		schm,  							// service control manager handle
		schs   : SC_Handle;  			// service handle
		ss     : TServiceStatus;		// service status
		dwStat : Integer;				// current service status
	begin
		dwStat := -1;

	  	// connect to the service control manager
		schm := OpenSCManager(PChar(sMachine), Nil, SC_MANAGER_CONNECT);

	  	// if successful...
		if(schm > 0)then
		begin
			// open a handle to the specified service
			// we want to query service status
			schs := OpenService(schm, PChar(sService), SERVICE_QUERY_STATUS);

			// if successful...
   		if(schs > 0)then
    		begin
      	// retrieve the current status of the specified service
      	if(QueryServiceStatus(schs, ss))then
				dwStat := ss.dwCurrentState;

			// close service handle
      	CloseServiceHandle(schs);
	    	end;

   	 	// close service control manager handle
    		CloseServiceHandle(schm);
  		end;
  		Result := dwStat;
	end;

	function ServiceRunning(sMachine, sService : string ) : boolean;
	// Purpose : return TRUE if the specified service is running, defined by
   // 			the status code SERVICE_RUNNING. Return FALSE if the service
	// 			is in any other state, including any pending states
	// Created	: http://www.chami.com/tips/delphi/031498D.html

	begin
		Result := SERVICE_RUNNING = ServiceGetStatus(sMachine, sService );
	end;


	function ServiceStopped(sMachine,sService : string ) : boolean;
	// Purpose	: return TRUE if the specified service was stopped,
   //					defined by the status code SERVICE_STOPPED.
	// Created	: http://www.chami.com/tips/delphi/031498D.html
	begin
		Result := SERVICE_STOPPED = ServiceGetStatus(sMachine, sService );
	end;

	Function ProcessRunning(sName: string) : boolean;
	// Purpose	: return TRUE if the specified process is running,
	// Created	: YR - 28.11.2000. Using JEDI Code Library beta1
   // Updated  : 1.04.3 : Fix a bug for W2k. W2K SP1 use full path to process.
	var
		ProcList : TStringList;
      i : integer;
      sProcess : string;
	begin
		ProcessRunning:=False;
	   ProcList := TStringList.Create;
		if RunningProcessesList(ProcList) then
      begin
			for i:=0 to ProcList.count-1 do
         begin
      		sProcess := copy(ProcList[i], lastdelimiter('\',ProcList[i])+1,
				                  length(ProcList[i]) - lastdelimiter('\',ProcList[i]) + 1);

				// Process name are not equal in term of case between
            // process table and Windows NT Task manager. So let's forget about
            // case sensitivity.
            if (StrIComp(PChar(sName),PChar(sProcess))=0) then
		     			ProcessRunning := True
	      end
	   end;
	end;

	procedure WriteToEventLog(sMsg : String);
	// Purpose	: Write an event in the NT event log.
	// Created	: YR - 28.11.2000
	var
	  	hEventLog : THandle;   // Handle to the event log.
	begin
	  	hEventLog := RegisterEventSource(nil, 'NSClient');

	  	if (hEventLog <= 0) then begin
	    	raise Exception.Create('RegisterEventSource call failed!');
	   	Exit;
	  	end;

		if not ReportEvent(hEventLog, 0, 0, 0, nil, 1, 0, @sMsg, nil) then begin
	   	DeregisterEventSource(hEventLog);
			raise Exception.Create('Failure in Report Event');
		end;

		DeregisterEventSource(hEventLog);
	end;

end.
