unit uNSClient;
// Purpose : Implementation of the service.
// Created : YR - November 2000
// Updated : 1.04.2 - 26.02.2001 YR - Add the "Custom Counter" feature
//           1.04.3 - 28.02.2001 YR - Process check was incorrect under W2K SP1. Fixed.
//           1.04.4 - 05.03.2001 YR - CUSTOM counters number are now return with two digits precision
 
interface

uses Windows, Messages, SysUtils, Classes, Graphics, Controls, SvcMgr, Dialogs,
		libSystem, pdh, pdhmsg, ExtCtrls, ScktComp, Registry, JclFileUtils, JclSYSInfo;

	// Main class
	type TNSClient = class(TService)
      ServerSocket: TServerSocket;
		CPUTimer: TTimer;
    	procedure ServerSocketClientRead(Sender: TObject; Socket: TCustomWinSocket);
      Procedure SetArgvFromRequest(sRequest : string);
     	procedure ServiceStart(Sender: TService; var Started: Boolean);
    	procedure CPUTimerTimer(Sender: TObject);
		procedure ServiceAfterUninstall(Sender: TService);
    	procedure ServiceAfterInstall(Sender: TService);
    procedure ServiceStop(Sender: TService; var Stopped: Boolean);
		private
         verInfo : TJclFileVersionInfo;
			sPassword : string;
			query : HQUERY;
	   	cpu_counter : HCOUNTER;
         uptime_counter : HCOUNTER;
         mem_commitLimit_counter : HCOUNTER;
         mem_commitByte_counter : HCOUNTER;
        	CounterValue: TPdhFmtCounterValue;
         CPUBuffer : TCircularBuffer;
         lnUptime : Int64;
         lnMemCommitLimit : Int64;
         lnMemCommitByte: Int64;
         asArgv : array of string;
         Function PDHCheck(sCounter: string; const Error: Cardinal): Longint;
         Function PdhGetErrorMsg(sCounter : string; const Error: Cardinal) : string;
     		Procedure CollectData;

         Function FindAnswer(sRequest: string) : String;
		   Function ProcessServReq: string;
		   Function ProcessProcReq: string;
		   Function ProcessCounterReq(sCounter : string): string;
 		public
    		function GetServiceController: TServiceController; override;
		published

    	{ Public declarations }
	end;

  	//Record to hold the information from one thread
  	TThreadInfo = record
		Active       : Boolean;
     	ThreadHandle : integer;
     	ThreadId     : Cardinal;
     	CurrentFile  : string;
  	end;

	const PRG_VERSION = 'NSClient V1.04 - 22.02.2001';
   // Requests constants declaration
	const REQ_VERSION = 1;
	const REQ_CPUTOTAL = 2;
	const REQ_UPTIME = 3;
   const REQ_FREESPACE = 4;
   const REQ_SERVICESTATE = 5;
   const REQ_RUNNINGPROCESS = 6;
	const REQ_MEMORYUSE = 7;
   const REQ_COUNTER = 8 ;
var
  NSClient: TNSClient;

implementation

{$R *.DFM}

	function TNSClient.PdhCheck(sCounter : string; const Error: Cardinal) : Longint;
	// Purpose	: Check wrapper for all PDH queries
	begin
      PdhCheck := Error;
      if Error <> ERROR_SUCCESS then
         WriteToEventLog('NSClient Error: PDH.dll ' + PdhGetErrorMsg(sCounter, Error))
	end;

   function TNSClient.PdhGetErrorMsg(sCounter : string; const Error: Cardinal) : string;
	// Purpose	: Return a string corresponding to the error
	// Created	: YR - 23.02.2001
   begin
      case Error of
         PDH_CSTATUS_ITEM_NOT_VALIDATED: PdhGetErrorMsg := sCounter + ' - ERROR: ITEM_NOT_VALIDATED';
      else
         PdhGetErrorMsg := sCounter + ' - ERROR: 0x' + IntToHex(Error,8)
      end
   end;

	procedure ServiceController(CtrlCode: DWord); stdcall;
	begin
	  NSClient.Controller(CtrlCode);
	end;

	function TNSClient.GetServiceController: TServiceController;
	begin
	  Result := ServiceController;
	end;

	procedure TNSClient.ServiceStart(Sender: TService; var Started: Boolean);
	// Purpose	: Initialize the counters
	// Created	: YR - 21.11.2000
	var
	  Reg: TRegistry;
	begin
		try
			CPUBuffer := TCircularBuffer.Create(17280);
         verInfo := TJclFileVersionInfo.Create(ParamStr(0));
			// Read config informations in the registry
	  		Reg := TRegistry.Create;
			Reg.RootKey:= HKEY_LOCAL_MACHINE;
	  		if not Reg.KeyExists('\Software\NSClient\Params') then
	      begin
		      ServerSocket.port := 1248;
	         sPassword := 'None';
	      end
	   	else
	      begin
				Reg.OpenKey('\Software\NSClient\Params',true);
	         ServerSocket.port := Reg.ReadInteger('port');
	      	sPassword := Reg.ReadString('password');
	      end;
	    	Reg.CloseKey;
	    	Reg.Free;

			// Start listen to the port.
	  		ServerSocket.Active:=True;

			// Create the PDH handle to collect CPU and uptime informations
		 	PdhCheck('OpenQuery Standard', pdhOpenQuery(nil,0,query));

	  		// Initialize the CPU Counter query
			if GetWindowsVersion in [wvWinNT4] then	// Windows NT 4 only
			   PdhCheck('\System\% Total Processor Time', pdhAddCounter(query,PChar('\\' + Get_ComputerName +
			   	'\System\% Total Processor Time'),0,cpu_counter))
	      else
		   	PdhCheck('\Processor(_total)\% Processor Time', pdhAddCounter(query,PChar('\\' + Get_ComputerName +
		   		'\Processor(_total)\% Processor Time'),0,cpu_counter));

	  		// Initialize the uptime counter query
		   PdhCheck('\System\System Up Time', pdhAddCounter(query,PChar('\\' + Get_ComputerName +
		   	'\System\System Up Time'),0,uptime_counter));
	      // Initialize the memory counters
			PdhCheck('\Memory\Commit Limit', pdhAddCounter(query,PChar('\\' + Get_ComputerName +
		   	'\Memory\Commit Limit'),0,mem_commitLimit_counter));
			PdhCheck('\Memory\Committed Bytes', pdhAddCounter(query,PChar('\\' + Get_ComputerName +
		   	'\Memory\Committed Bytes'),0,mem_commitByte_counter));

         CPUTimer.Enabled := true;
      	WriteToEventLog('NSClient has started');
   	except
      	WriteToEventLog('Service start failed');
      end;
   end;

	Procedure TNSClient.CollectData;
	// Purpose	: Get CPU load and add it to a circular buffer
	// Created	: YR - 21.11.2000
	begin
		// Get the informations
		PdhCheck('Collect', PdhCollectQueryData(query));

      // Process CPU counter
		PdhCheck('Collect CPU', PdhGetFormattedCounterValue(cpu_counter, PDH_FMT_LARGE, nil, CounterValue));
		case CounterValue.CStatus of
			PDH_CSTATUS_NEW_DATA: CPUBuffer.Push(CounterValue.LargeValue);
		   PDH_CSTATUS_VALID_DATA: CPUBuffer.Push(CounterValue.LargeValue);
	   end;

      // Process Uptime counter
		PdhCheck('Collect Uptime',PdhGetFormattedCounterValue(uptime_counter, PDH_FMT_LARGE, nil, CounterValue));
		case CounterValue.CStatus of
			PDH_CSTATUS_NEW_DATA: lnUptime:=CounterValue.LargeValue;
		   PDH_CSTATUS_VALID_DATA: lnUptime:=CounterValue.LargeValue;
	   end;

      // Process memory counters
		PdhCheck('Collect CommitLimit',PdhGetFormattedCounterValue(mem_commitLimit_counter, PDH_FMT_LARGE, nil, CounterValue));
		case CounterValue.CStatus of
			PDH_CSTATUS_NEW_DATA, PDH_CSTATUS_VALID_DATA: lnMemCommitLimit:=CounterValue.LargeValue;
	   end;

		PdhCheck('Collect CommitByte',PdhGetFormattedCounterValue(mem_commitByte_counter, PDH_FMT_LARGE, nil, CounterValue));
		case CounterValue.CStatus of
			PDH_CSTATUS_NEW_DATA, PDH_CSTATUS_VALID_DATA: lnMemCommitByte :=CounterValue.LargeValue;
	   end;
	end;

   Function TNSClient.FindAnswer(sRequest: string) : String;
	// Purpose	: Process all requests from the plugin
	// Created	: YR - 22.11.2000
	begin
      SetArgvFromRequest(sRequest);	// split the string into a string array
      try
			if sPassword = asArgv[0] then
				begin
					// If the first argument is not a number, Delphi will generate an exception
			      case StrToIntDef(asArgv[1],0) of
                  REQ_VERSION: FindAnswer := verInfo.BinFileVersion;
		         	REQ_CPUTOTAL: FindAnswer:= IntToStr(CPUBuffer.GetAverage(StrToIntDef(asArgv[2],60)*12));
  				      REQ_UPTIME: FindAnswer:= IntToStr(lnUptime);
						REQ_FREESPACE: FindAnswer:=IntToStr(GetFreeSpace(strupper(PChar(asArgv[2])))) + '&' +
                  	IntToStr(GetDiskSize(strupper(PChar(asArgv[2]))));
  				      REQ_SERVICESTATE: FindAnswer:= ProcessServReq;
                  REQ_RUNNINGPROCESS: FindAnswer := ProcessProcReq;
                  REQ_MEMORYUSE: FindAnswer := IntToStr(lnMemCommitLimit) + '&' + IntToStr(lnMemCommitByte);
                  REQ_COUNTER: FindAnswer := ProcessCounterReq(asArgv[2]);
		         else
		     			FindAnswer:= 'ERROR:Unknown request number';
			      end;
            end
         else
				FindAnswer:= 'ERROR:Wrong password';
      except
      	on E: Exception do
         	begin
         	FindAnswer := 'ERROR:Malformed request or internal error. Check EventLog:' +sRequest;
            WriteToEventLog(E.Message);
            end;
      end;
   end;

   Procedure TNSClient.SetArgvFromRequest(sRequest : string);
	// Purpose	: Set the asArgv array according to the request we received
	// Created	: YR - 22.11.2000
	var
      nPos : integer;
      nArgs : integer;
      i: integer;
   begin
		SetLength(asArgv,10);
		for i:=0 to length(asArgv)-1 do asArgv[i]:='';	// Fill the params array with ''

		// Fill the array
      nArgs:=0;
		nPos := Pos('&',sRequest);
		while nPos>0 do begin
      	asArgv[nArgs] := copy(sRequest,1,nPos-1);		// Get the parameter
         sRequest := copy(sRequest, nPos+1, maxint);	// Remove the parameter from the request
         inc(nArgs);             							// inc number of parameters
         if nArgs>=length(asArgv) then    				// increase array if too small
         	setlength(asArgv,length(asArgv)+10);
     		nPos := Pos('&',sRequest);							// look for another parameters
      end;
		asArgv[nArgs]:=sRequest;
		SetLength(asArgv,nArgs+1);								// set the length of the array
   end;

	procedure TNSClient.ServerSocketClientRead(Sender: TObject; Socket: TCustomWinSocket);
	// Purpose	: Receive the TCP requests and send the answer to the plugin
	// Created	: YR - 22.11.2000
	begin
		Socket.SendText(FindAnswer(Socket.ReceiveText));
	end;

	procedure TNSClient.CPUTimerTimer(Sender: TObject);
   // Purpose	: Collect CPU load data
	// Created	: YR - 22.11.2000
	begin
		try
			CollectData;
      except 
      	WriteToEventLog('CollectData failed');
      end;
	end;

   function TNSClient.ProcessServReq: string;
   // Purpose	: Check the state of all services and return a status and
   // 				the faulty services
	// Created	: YR - 27.11.2000
	var
   	i : integer;			// loop's index
      s : string;				// answer
      nState : integer;		// Overall service state
      bShowAll : boolean;	// true is we have to display started services too
   begin
		nState := 0;							// OK State
      s := '';
      bShowAll := False;
		if length(asArgv) >= 4 then	// minimum number of parameters
      begin
			// Do we have to return started services too ?
			if strupper(PChar(asArgv[2])) = 'SHOWALL' then
         	bShowAll:=true;

			// Loop on every service we have to check
	      for i:=3 to length(asArgv)-1 do
				if ServiceRunning('',asArgv[i]) then begin
            	// Service is running
					if bShowAll=True then s := s + ' ' + asArgv[i] + ': Started -'
           	end
            else if (ServiceStopped('',asArgv[i])) then	begin
					// Service is stopped
					s := s + ' ' + asArgv[i] +': Stopped -';
              	nState := 2;
           	end
            else begin
					// service is unknown
            	s := s + ' ' + asArgv[i] +': Unknown -';
               if (nState<1) then nState := 1;
            end;
	      ;

			if length(s)=0 then
      		s := 'All services are running'		// if everything's fine
         else
				SetLength(s,length(s)-1);				// remove trailing -
		end
		else
      begin
      	// not enough parameters. Should never occur, but...
      	s := 'No service to check';
         nState := -1;
      end;
  	   ProcessServReq := IntToStr(nState) + '&' + s;
   end;

   function TNSClient.ProcessProcReq: string;
   // Purpose	: Check if all process are running and
   //					return a status and the faulty processes
	// Created	: YR - 28.11.2000
	var
   	i : integer;			// loop's index
      s : string;				// answer
      nState : integer;		// Overall Process state
      bShowAll : boolean;	// true is we have to display all processes
   begin
		nState := 0;							// OK State
      s := '';
      bShowAll := False;
		if length(asArgv) >= 4 then // minimum number of parameters
      begin
			// Do we have to return running process too ?
			if strupper(PChar(asArgv[2])) = 'SHOWALL' then
         	bShowAll:=true;

			// Loop on every process we have to check
	      for i:=3 to length(asArgv)-1 do
				if ProcessRunning(asArgv[i]) then
            begin
            	// Process is running
					if bShowAll=True then s := s + ' ' + asArgv[i] + ': Running -'
           	end
            else
            begin
					// Process is not running
            	s := s + ' ' + asArgv[i] +': not running -';
               nState := 2;
            end;
	      ;
			if length(s)=0 then
      		s := 'All processes are running'		// if everything's fine
         else
				SetLength(s,length(s)-1);				// remove trailing -
		end
		else begin
      	// not enough parameters
      	s := 'No process to check';
         nState := -1;
      end;
  	   ProcessProcReq := IntToStr(nState) + '&' + s;
   end;

   function TNSClient.ProcessCounterReq(sCounter : string) : string;
   // Purpose	: Get information about any counter. Just be sure it exists!
	// Created	: YR - 22.02.2001
   var
      qryPDHQuery : HQUERY;
      custom_counter : HCOUNTER;
      lStatus : cardinal;
   begin

//      WriteToEventLog('NSClient DEBUG: Custom Counter : ' + sCounter);

      // Open a new query
      PdhCheck('OpenQuery' + sCounter, pdhOpenQuery(nil,0,qryPDHQuery));

      // Add the counter
      lStatus := PdhCheck('Add ' + sCounter,pdhAddCounter(qryPDHQuery,PChar('\\' + Get_ComputerName + sCounter),0,custom_counter));

      if lStatus <> ERROR_SUCCESS then
         // Counter is invalid
         ProcessCounterReq := PdhGetErrorMsg(sCounter, lStatus)
      else
      begin
         // Counter is valid

         // Collect the data
         PdhCheck('Collect ' + sCounter, PdhCollectQueryData(qryPDHQuery));

         // Process Custom counter
      	PdhCheck(sCounter, PdhGetFormattedCounterValue(custom_counter, PDH_FMT_DOUBLE, nil, CounterValue));
      	case CounterValue.CStatus of
      	   PDH_CSTATUS_NEW_DATA, PDH_CSTATUS_VALID_DATA : ProcessCounterReq :=
               FloatToStrF(CounterValue.doubleValue,ffFixed,15,5);
         else
            ProcessCounterReq := PdhGetErrorMsg(sCounter, Error)
         end;
      end;

      // Close the query
      pdhCheck('Close ' + sCounter, pdhCloseQuery(qryPDHQuery));
   end;

	procedure TNSClient.ServiceAfterUninstall(Sender: TService);
   // Purpose	: Remove the registry entries after the servie is uninstalled
	// Created	: YR - 28.11.2000
	var
		Reg: TRegistry;
	begin
  		Reg := TRegistry.Create;
		Reg.RootKey:= HKEY_LOCAL_MACHINE;

		if Reg.KeyExists('\Software\NSClient\Params') then
	   begin
			Reg.DeleteValue('port');
  			Reg.DeleteValue('password');
			Reg.DeleteKey('\Software\NSClient\Params');
			Reg.DeleteKey('\Software\NSClient');
	   end;
  		WriteToEventLog('NSClient is uninstalled');
	end;

	procedure TNSClient.ServiceAfterInstall(Sender: TService);
   // Purpose	: Add the registry entries after the service is installed
	// Created	: YR - 28.11.2000
	var
		Reg: TRegistry;
	begin
		try
	  		Reg := TRegistry.Create;
			Reg.RootKey:= HKEY_LOCAL_MACHINE;

			if not Reg.KeyExists('\Software\NSClient\Params') then
		   begin
				// if the key doesn't exist, let's create it.
				Reg.CreateKey('\Software\NSClient\Params');
				Reg.OpenKey('\Software\NSClient\Params',true);
		      Reg.WriteInteger ('port',1248);
				Reg.WriteString ('password','None');
		   end;
	  		WriteToEventLog('NSClient installed');
      except
			WriteToEventLog('Service Afterinstall failed');
      end;
	end;

procedure TNSClient.ServiceStop(Sender: TService; var Stopped: Boolean);
   // Purpose	: Release the PDH query
	// Created	: YR - 23.02.2001

begin
   CPUTimer.Enabled := false;
   PdhCheck('Close standard', pdhCloseQuery(query));
	WriteToEventLog('NSClient has stopped');
end;

end.
